@@ -1,16 +1,17 @@
 use std::fmt::{self, Formatter};
-use std::os::raw::c_ulong;
 
-use glutin::window::Fullscreen;
-use log::error;
+use log::{error, warn};
 use serde::de::{self, MapAccess, Visitor};
 use serde::{Deserialize, Deserializer, Serialize};
+use winit::window::{Fullscreen, Theme};
 
-use alacritty_config_derive::ConfigDeserialize;
-use alacritty_terminal::config::{Percentage, LOG_TARGET_CONFIG};
-use alacritty_terminal::index::Column;
+#[cfg(target_os = "macos")]
+use winit::platform::macos::OptionAsAlt as WinitOptionAsAlt;
 
-use crate::config::ui_config::Delta;
+use alacritty_config_derive::{ConfigDeserialize, SerdeReplace};
+
+use crate::config::ui_config::{Delta, Percentage};
+use crate::config::LOG_TARGET_CONFIG;
 
 /// Default Alacritty name, used for window title and class.
 pub const DEFAULT_NAME: &str = "Alacritty";
@@ -28,10 +29,10 @@ pub struct WindowConfig {
 
     /// XEmbed parent.
     #[config(skip)]
-    pub embed: Option<c_ulong>,
+    pub embed: Option<u32>,
 
-    /// GTK theme variant.
-    pub gtk_theme_variant: Option<String>,
+    /// System decorations theme variant.
+    pub decorations_theme_variant: Option<Theme>,
 
     /// Spread out additional padding evenly.
     pub dynamic_padding: bool,
@@ -46,8 +47,18 @@ pub struct WindowConfig {
     /// Background opacity from 0.0 to 1.0.
     pub opacity: Percentage,
 
+    /// Request blur behind the window.
+    pub blur: bool,
+
+    /// Controls which `Option` key should be treated as `Alt`.
+    #[cfg(target_os = "macos")]
+    option_as_alt: OptionAsAlt,
+
+    /// Resize increments.
+    pub resize_increments: bool,
+
     /// Pixel padding.
-    padding: Delta<u8>,
+    padding: Delta<u16>,
 
     /// Initial dimensions.
     dimensions: Dimensions,
@@ -57,16 +68,20 @@ impl Default for WindowConfig {
     fn default() -> Self {
         Self {
             dynamic_title: true,
+            blur: Default::default(),
+            embed: Default::default(),
+            padding: Default::default(),
+            opacity: Default::default(),
             position: Default::default(),
+            identity: Default::default(),
+            dimensions: Default::default(),
             decorations: Default::default(),
             startup_mode: Default::default(),
-            embed: Default::default(),
-            gtk_theme_variant: Default::default(),
             dynamic_padding: Default::default(),
-            identity: Identity::default(),
-            opacity: Default::default(),
-            padding: Default::default(),
-            dimensions: Default::default(),
+            resize_increments: Default::default(),
+            decorations_theme_variant: Default::default(),
+            #[cfg(target_os = "macos")]
+            option_as_alt: Default::default(),
         }
     }
 }
@@ -74,20 +89,40 @@ impl Default for WindowConfig {
 impl WindowConfig {
     #[inline]
     pub fn dimensions(&self) -> Option<Dimensions> {
-        if self.dimensions.columns.0 != 0
-            && self.dimensions.lines != 0
-            && self.startup_mode != StartupMode::Maximized
-        {
+        let (lines, columns) = (self.dimensions.lines, self.dimensions.columns);
+        let (lines_is_non_zero, columns_is_non_zero) = (lines != 0, columns != 0);
+
+        if lines_is_non_zero && columns_is_non_zero {
+            // Return dimensions if both `lines` and `columns` are non-zero.
             Some(self.dimensions)
+        } else if lines_is_non_zero || columns_is_non_zero {
+            // Warn if either `columns` or `lines` is non-zero.
+
+            let (zero_key, non_zero_key, non_zero_value) = if lines_is_non_zero {
+                ("columns", "lines", lines)
+            } else {
+                ("lines", "columns", columns)
+            };
+
+            warn!(
+                target: LOG_TARGET_CONFIG,
+                "Both `lines` and `columns` must be non-zero for `window.dimensions` to take \
+                 effect. Configured value of `{}` is 0 while that of `{}` is {}",
+                zero_key,
+                non_zero_key,
+                non_zero_value,
+            );
+
+            None
         } else {
             None
         }
     }
 
     #[inline]
-    pub fn padding(&self, dpr: f64) -> (f32, f32) {
-        let padding_x = (f32::from(self.padding.x) * dpr as f32).floor();
-        let padding_y = (f32::from(self.padding.y) * dpr as f32).floor();
+    pub fn padding(&self, scale_factor: f32) -> (f32, f32) {
+        let padding_x = (f32::from(self.padding.x) * scale_factor).floor();
+        let padding_y = (f32::from(self.padding.y) * scale_factor).floor();
         (padding_x, padding_y)
     }
 
@@ -104,9 +139,19 @@ impl WindowConfig {
     pub fn maximized(&self) -> bool {
         self.startup_mode == StartupMode::Maximized
     }
+
+    #[cfg(target_os = "macos")]
+    pub fn option_as_alt(&self) -> WinitOptionAsAlt {
+        match self.option_as_alt {
+            OptionAsAlt::OnlyLeft => WinitOptionAsAlt::OnlyLeft,
+            OptionAsAlt::OnlyRight => WinitOptionAsAlt::OnlyRight,
+            OptionAsAlt::Both => WinitOptionAsAlt::Both,
+            OptionAsAlt::None => WinitOptionAsAlt::None,
+        }
+    }
 }
 
-#[derive(ConfigDeserialize, Debug, Clone, PartialEq)]
+#[derive(ConfigDeserialize, Debug, Clone, PartialEq, Eq)]
 pub struct Identity {
     /// Window title.
     pub title: String,
@@ -121,8 +166,9 @@ impl Default for Identity {
     }
 }
 
-#[derive(ConfigDeserialize, Debug, Copy, Clone, PartialEq, Eq)]
+#[derive(ConfigDeserialize, Default, Debug, Copy, Clone, PartialEq, Eq)]
 pub enum StartupMode {
+    #[default]
     Windowed,
     Maximized,
     Fullscreen,
@@ -130,14 +176,9 @@ pub enum StartupMode {
     SimpleFullscreen,
 }
 
-impl Default for StartupMode {
-    fn default() -> StartupMode {
-        StartupMode::Windowed
-    }
-}
-
-#[derive(ConfigDeserialize, Debug, Copy, Clone, PartialEq, Eq)]
+#[derive(ConfigDeserialize, Default, Debug, Copy, Clone, PartialEq, Eq)]
 pub enum Decorations {
+    #[default]
     Full,
     #[cfg(target_os = "macos")]
     Transparent,
@@ -146,34 +187,34 @@ pub enum Decorations {
     None,
 }
 
-impl Default for Decorations {
-    fn default() -> Decorations {
-        Decorations::Full
-    }
-}
-
 /// Window Dimensions.
 ///
 /// Newtype to avoid passing values incorrectly.
 #[derive(ConfigDeserialize, Default, Debug, Copy, Clone, PartialEq, Eq)]
 pub struct Dimensions {
     /// Window width in character columns.
-    pub columns: Column,
+    pub columns: usize,
 
     /// Window Height in character lines.
     pub lines: usize,
 }
 
 /// Window class hint.
-#[derive(Serialize, Debug, Clone, PartialEq, Eq)]
+#[derive(SerdeReplace, Serialize, Debug, Clone, PartialEq, Eq)]
 pub struct Class {
-    pub instance: String,
     pub general: String,
+    pub instance: String,
+}
+
+impl Class {
+    pub fn new(general: impl ToString, instance: impl ToString) -> Self {
+        Self { general: general.to_string(), instance: instance.to_string() }
+    }
 }
 
 impl Default for Class {
     fn default() -> Self {
-        Self { instance: DEFAULT_NAME.into(), general: DEFAULT_NAME.into() }
+        Self::new(DEFAULT_NAME, DEFAULT_NAME)
     }
 }
 
@@ -203,7 +244,7 @@ impl<'de> Deserialize<'de> for Class {
             {
                 let mut class = Self::Value::default();
 
-                while let Some((key, value)) = map.next_entry::<String, serde_yaml::Value>()? {
+                while let Some((key, value)) = map.next_entry::<String, toml::Value>()? {
                     match key.as_str() {
                         "instance" => match String::deserialize(value) {
                             Ok(instance) => class.instance = instance,
@@ -223,7 +264,7 @@ impl<'de> Deserialize<'de> for Class {
                                 );
                             },
                         },
-                        _ => (),
+                        key => warn!(target: LOG_TARGET_CONFIG, "Unrecognized class field: {key}"),
                     }
                 }
 
@@ -234,3 +275,20 @@ impl<'de> Deserialize<'de> for Class {
         deserializer.deserialize_any(ClassVisitor)
     }
 }
+
+#[cfg(target_os = "macos")]
+#[derive(ConfigDeserialize, Default, Debug, Clone, Copy, PartialEq, Eq)]
+pub enum OptionAsAlt {
+    /// The left `Option` key is treated as `Alt`.
+    OnlyLeft,
+
+    /// The right `Option` key is treated as `Alt`.
+    OnlyRight,
+
+    /// Both `Option` keys are treated as `Alt`.
+    Both,
+
+    /// No special handling is applied for `Option` key.
+    #[default]
+    None,
+}
