@@ -1,16 +1,21 @@
-use std::cmp::{max, min};
+use std::cmp::Reverse;
+use std::collections::HashSet;
+use std::iter;
 
-use glutin::event::ModifiersState;
+use ahash::RandomState;
+use winit::keyboard::ModifiersState;
 
-use alacritty_terminal::grid::BidirectionalIterator;
-use alacritty_terminal::index::{Boundary, Direction, Point};
+use alacritty_terminal::grid::{BidirectionalIterator, Dimensions};
+use alacritty_terminal::index::{Boundary, Column, Direction, Line, Point};
+use alacritty_terminal::term::cell::Hyperlink;
 use alacritty_terminal::term::search::{Match, RegexIter, RegexSearch};
 use alacritty_terminal::term::{Term, TermMode};
 
 use crate::config::ui_config::{Hint, HintAction};
-use crate::config::Config;
-use crate::display::content::RegexMatches;
-use crate::display::MAX_SEARCH_LINES;
+use crate::config::UiConfig;
+
+/// Maximum number of linewraps followed outside of the viewport during search highlighting.
+pub const MAX_SEARCH_LINES: usize = 100;
 
 /// Percentage of characters in the hints alphabet used for the last character.
 const HINT_SPLIT_PERCENTAGE: f32 = 0.5;
@@ -24,7 +29,7 @@ pub struct HintState {
     alphabet: String,
 
     /// Visible matches.
-    matches: RegexMatches,
+    matches: Vec<Match>,
 
     /// Key label for each visible match.
     labels: Vec<Vec<char>>,
@@ -70,21 +75,30 @@ impl HintState {
             None => return,
         };
 
-        // Find visible matches.
-        self.matches.0 = hint.regex.with_compiled(|regex| {
-            let mut matches = RegexMatches::new(term, regex);
+        // Clear current matches.
+        self.matches.clear();
 
-            // Apply post-processing and search for sub-matches if necessary.
-            if hint.post_processing {
-                matches
-                    .drain(..)
-                    .map(|rm| HintPostProcessor::new(term, regex, rm).collect::<Vec<_>>())
-                    .flatten()
-                    .collect()
-            } else {
-                matches.0
-            }
-        });
+        // Add escape sequence hyperlinks.
+        if hint.content.hyperlinks {
+            self.matches.extend(visible_unique_hyperlinks_iter(term));
+        }
+
+        // Add visible regex matches.
+        if let Some(regex) = hint.content.regex.as_ref() {
+            regex.with_compiled(|regex| {
+                let matches = visible_regex_match_iter(term, regex);
+
+                // Apply post-processing and search for sub-matches if necessary.
+                if hint.post_processing {
+                    let mut matches = matches.collect::<Vec<_>>();
+                    self.matches.extend(matches.drain(..).flat_map(|rm| {
+                        HintPostProcessor::new(term, regex, rm).collect::<Vec<_>>()
+                    }));
+                } else {
+                    self.matches.extend(matches);
+                }
+            });
+        }
 
         // Cancel highlight with no visible matches.
         if self.matches.is_empty() {
@@ -92,6 +106,10 @@ impl HintState {
             return;
         }
 
+        // Sort and dedup ranges. Currently overlapped but not exactly same ranges are kept.
+        self.matches.sort_by_key(|bounds| (*bounds.start(), Reverse(*bounds.end())));
+        self.matches.dedup_by_key(|bounds| *bounds.start());
+
         let mut generator = HintLabels::new(&self.alphabet, HINT_SPLIT_PERCENTAGE);
         let match_count = self.matches.len();
         let keys_len = self.keys.len();
@@ -134,9 +152,16 @@ impl HintState {
             let bounds = self.matches[index].clone();
             let action = hint.action.clone();
 
-            self.stop();
+            // Exit hint mode unless it requires explicit dismissal.
+            if hint.persist {
+                self.keys.clear();
+            } else {
+                self.stop();
+            }
 
-            Some(HintMatch { action, bounds })
+            // Hyperlinks take precedence over regex matches.
+            let hyperlink = term.grid()[*bounds.start()].hyperlink();
+            Some(HintMatch { action, bounds, hyperlink })
         } else {
             // Store character to preserve the selection.
             self.keys.push(c);
@@ -151,7 +176,7 @@ impl HintState {
     }
 
     /// Visible hint regex matches.
-    pub fn matches(&self) -> &RegexMatches {
+    pub fn matches(&self) -> &[Match] {
         &self.matches
     }
 
@@ -165,13 +190,36 @@ impl HintState {
 }
 
 /// Hint match which was selected by the user.
-#[derive(PartialEq, Debug, Clone)]
+#[derive(PartialEq, Eq, Debug, Clone)]
 pub struct HintMatch {
     /// Action for handling the text.
-    pub action: HintAction,
+    action: HintAction,
 
     /// Terminal range matching the hint.
-    pub bounds: Match,
+    bounds: Match,
+
+    hyperlink: Option<Hyperlink>,
+}
+
+impl HintMatch {
+    #[inline]
+    pub fn should_highlight(&self, point: Point, pointed_hyperlink: Option<&Hyperlink>) -> bool {
+        self.bounds.contains(&point) && self.hyperlink.as_ref() == pointed_hyperlink
+    }
+
+    #[inline]
+    pub fn action(&self) -> &HintAction {
+        &self.action
+    }
+
+    #[inline]
+    pub fn bounds(&self) -> &Match {
+        &self.bounds
+    }
+
+    pub fn hyperlink(&self) -> Option<&Hyperlink> {
+        self.hyperlink.as_ref()
+    }
 }
 
 /// Generator for creating new hint labels.
@@ -239,16 +287,87 @@ impl HintLabels {
     }
 }
 
+/// Iterate over all visible regex matches.
+pub fn visible_regex_match_iter<'a, T>(
+    term: &'a Term<T>,
+    regex: &'a mut RegexSearch,
+) -> impl Iterator<Item = Match> + 'a {
+    let viewport_start = Line(-(term.grid().display_offset() as i32));
+    let viewport_end = viewport_start + term.bottommost_line();
+    let mut start = term.line_search_left(Point::new(viewport_start, Column(0)));
+    let mut end = term.line_search_right(Point::new(viewport_end, Column(0)));
+    start.line = start.line.max(viewport_start - MAX_SEARCH_LINES);
+    end.line = end.line.min(viewport_end + MAX_SEARCH_LINES);
+
+    RegexIter::new(start, end, Direction::Right, term, regex)
+        .skip_while(move |rm| rm.end().line < viewport_start)
+        .take_while(move |rm| rm.start().line <= viewport_end)
+}
+
+/// Iterate over all visible hyperlinks, yanking only unique ones.
+pub fn visible_unique_hyperlinks_iter<T>(term: &Term<T>) -> impl Iterator<Item = Match> + '_ {
+    let mut display_iter = term.grid().display_iter().peekable();
+
+    // Avoid creating hints for the same hyperlinks, but from a different places.
+    let mut unique_hyperlinks = HashSet::<Hyperlink, RandomState>::default();
+
+    iter::from_fn(move || {
+        // Find the start of the next unique hyperlink.
+        let (cell, hyperlink) = display_iter.find_map(|cell| {
+            let hyperlink = cell.hyperlink()?;
+            (!unique_hyperlinks.contains(&hyperlink)).then(|| {
+                unique_hyperlinks.insert(hyperlink.clone());
+                (cell, hyperlink)
+            })
+        })?;
+
+        let start = cell.point;
+        let mut end = start;
+
+        // Find the end bound of just found unique hyperlink.
+        while let Some(next_cell) = display_iter.peek() {
+            // Cell at display iter doesn't match, yield the hyperlink and start over with
+            // `find_map`.
+            if next_cell.hyperlink().as_ref() != Some(&hyperlink) {
+                break;
+            }
+
+            // Advance to the next cell.
+            end = next_cell.point;
+            let _ = display_iter.next();
+        }
+
+        Some(start..=end)
+    })
+}
+
+/// Retrieve the match, if the specified point is inside the content matching the regex.
+fn regex_match_at<T>(
+    term: &Term<T>,
+    point: Point,
+    regex: &mut RegexSearch,
+    post_processing: bool,
+) -> Option<Match> {
+    let regex_match = visible_regex_match_iter(term, regex).find(|rm| rm.contains(&point))?;
+
+    // Apply post-processing and search for sub-matches if necessary.
+    if post_processing {
+        HintPostProcessor::new(term, regex, regex_match).find(|rm| rm.contains(&point))
+    } else {
+        Some(regex_match)
+    }
+}
+
 /// Check if there is a hint highlighted at the specified point.
 pub fn highlighted_at<T>(
     term: &Term<T>,
-    config: &Config,
+    config: &UiConfig,
     point: Point,
     mouse_mods: ModifiersState,
 ) -> Option<HintMatch> {
     let mouse_mode = term.mode().intersects(TermMode::MOUSE_MODE);
 
-    config.ui_config.hints.enabled.iter().find_map(|hint| {
+    config.hints.enabled.iter().find_map(|hint| {
         // Check if all required modifiers are pressed.
         let highlight = hint.mouse.map_or(false, |mouse| {
             mouse.enabled
@@ -259,36 +378,80 @@ pub fn highlighted_at<T>(
             return None;
         }
 
-        hint.regex.with_compiled(|regex| {
-            // Setup search boundaries.
-            let mut start = term.line_search_left(point);
-            start.line = max(start.line, point.line - MAX_SEARCH_LINES);
-            let mut end = term.line_search_right(point);
-            end.line = min(end.line, point.line + MAX_SEARCH_LINES);
+        if let Some((hyperlink, bounds)) =
+            hint.content.hyperlinks.then(|| hyperlink_at(term, point)).flatten()
+        {
+            return Some(HintMatch {
+                bounds,
+                action: hint.action.clone(),
+                hyperlink: Some(hyperlink),
+            });
+        }
 
-            // Function to verify that the specified point is inside the match.
-            let at_point = |rm: &Match| *rm.end() >= point && *rm.start() <= point;
+        let bounds = hint.content.regex.as_ref().and_then(|regex| {
+            regex.with_compiled(|regex| regex_match_at(term, point, regex, hint.post_processing))
+        });
+        if let Some(bounds) = bounds.flatten() {
+            return Some(HintMatch { bounds, action: hint.action.clone(), hyperlink: None });
+        }
 
-            // Check if there's any match at the specified point.
-            let mut iter = RegexIter::new(start, end, Direction::Right, term, regex);
-            let regex_match = iter.find(at_point)?;
-
-            // Apply post-processing and search for sub-matches if necessary.
-            let regex_match = if hint.post_processing {
-                HintPostProcessor::new(term, regex, regex_match).find(at_point)
-            } else {
-                Some(regex_match)
-            };
-
-            regex_match.map(|bounds| HintMatch { action: hint.action.clone(), bounds })
-        })
+        None
     })
 }
 
+/// Retrieve the hyperlink with its range, if there is one at the specified point.
+fn hyperlink_at<T>(term: &Term<T>, point: Point) -> Option<(Hyperlink, Match)> {
+    let hyperlink = term.grid()[point].hyperlink()?;
+
+    let viewport_start = Line(-(term.grid().display_offset() as i32));
+    let viewport_end = viewport_start + term.bottommost_line();
+
+    let mut match_start = Point::new(point.line, Column(0));
+    let mut match_end = Point::new(point.line, Column(term.columns() - 1));
+    let grid = term.grid();
+
+    // Find adjacent lines that have the same `hyperlink`. The end purpose to highlight hyperlinks
+    // that span across multiple lines or not directly attached to each other.
+
+    // Find the closest to the viewport start adjacent line.
+    while match_start.line > viewport_start {
+        let next_line = match_start.line - 1i32;
+        // Iterate over all the cells in the grid's line and check if any of those cells contains
+        // the hyperlink we've found at original `point`.
+        let line_contains_hyperlink = grid[next_line]
+            .into_iter()
+            .any(|cell| cell.hyperlink().map_or(false, |h| h == hyperlink));
+
+        // There's no hyperlink on the next line, break.
+        if !line_contains_hyperlink {
+            break;
+        }
+
+        match_start.line = next_line;
+    }
+
+    // Ditto for the end.
+    while match_end.line < viewport_end {
+        let next_line = match_end.line + 1i32;
+
+        let line_contains_hyperlink = grid[next_line]
+            .into_iter()
+            .any(|cell| cell.hyperlink().map_or(false, |h| h == hyperlink));
+
+        if !line_contains_hyperlink {
+            break;
+        }
+
+        match_end.line = next_line;
+    }
+
+    Some((hyperlink, match_start..=match_end))
+}
+
 /// Iterator over all post-processed matches inside an existing hint match.
 struct HintPostProcessor<'a, T> {
     /// Regex search DFAs.
-    regex: &'a RegexSearch,
+    regex: &'a mut RegexSearch,
 
     /// Terminal reference.
     term: &'a Term<T>,
@@ -305,14 +468,17 @@ struct HintPostProcessor<'a, T> {
 
 impl<'a, T> HintPostProcessor<'a, T> {
     /// Create a new iterator for an unprocessed match.
-    fn new(term: &'a Term<T>, regex: &'a RegexSearch, regex_match: Match) -> Self {
-        let end = *regex_match.end();
-        let mut post_processor = Self { next_match: None, start: end, end, term, regex };
+    fn new(term: &'a Term<T>, regex: &'a mut RegexSearch, regex_match: Match) -> Self {
+        let mut post_processor = Self {
+            next_match: None,
+            start: *regex_match.start(),
+            end: *regex_match.end(),
+            term,
+            regex,
+        };
 
         // Post-process the first hint match.
-        let next_match = post_processor.hint_post_processing(&regex_match);
-        post_processor.start = next_match.end().add(term, Boundary::Grid, 1);
-        post_processor.next_match = Some(next_match);
+        post_processor.next_processed_match(regex_match);
 
         post_processor
     }
@@ -323,7 +489,7 @@ impl<'a, T> HintPostProcessor<'a, T> {
     /// to be unlikely to be intentionally part of the hint.
     ///
     /// This is most useful for identifying URLs appropriately.
-    fn hint_post_processing(&self, regex_match: &Match) -> Match {
+    fn hint_post_processing(&self, regex_match: &Match) -> Option<Match> {
         let mut iter = self.term.grid().iter_from(*regex_match.start());
 
         let mut c = iter.cell().c;
@@ -378,7 +544,31 @@ impl<'a, T> HintPostProcessor<'a, T> {
             }
         }
 
-        start..=iter.point()
+        if start > iter.point() {
+            None
+        } else {
+            Some(start..=iter.point())
+        }
+    }
+
+    /// Loop over submatches until a non-empty post-processed match is found.
+    fn next_processed_match(&mut self, mut regex_match: Match) {
+        self.next_match = loop {
+            if let Some(next_match) = self.hint_post_processing(&regex_match) {
+                self.start = next_match.end().add(self.term, Boundary::Grid, 1);
+                break Some(next_match);
+            }
+
+            self.start = regex_match.start().add(self.term, Boundary::Grid, 1);
+            if self.start > self.end {
+                return;
+            }
+
+            match self.term.regex_search_right(self.regex, self.start, self.end) {
+                Some(rm) => regex_match = rm,
+                None => return,
+            }
+        };
     }
 }
 
@@ -390,9 +580,7 @@ impl<'a, T> Iterator for HintPostProcessor<'a, T> {
 
         if self.start <= self.end {
             if let Some(rm) = self.term.regex_search_right(self.regex, self.start, self.end) {
-                let regex_match = self.hint_post_processing(&rm);
-                self.start = regex_match.end().add(self.term, Boundary::Grid, 1);
-                self.next_match = Some(regex_match);
+                self.next_processed_match(rm);
             }
         }
 
@@ -402,6 +590,10 @@ impl<'a, T> Iterator for HintPostProcessor<'a, T> {
 
 #[cfg(test)]
 mod tests {
+    use alacritty_terminal::index::{Column, Line};
+    use alacritty_terminal::term::test::mock_term;
+    use alacritty_terminal::vte::ansi::Handler;
+
     use super::*;
 
     #[test]
@@ -442,4 +634,70 @@ mod tests {
         assert_eq!(generator.next(), vec!['3', '3', '3', '0']);
         assert_eq!(generator.next(), vec!['3', '3', '3', '1']);
     }
+
+    #[test]
+    fn closed_bracket_does_not_result_in_infinite_iterator() {
+        let term = mock_term(" ) ");
+
+        let mut search = RegexSearch::new("[^/ ]").unwrap();
+
+        let count = HintPostProcessor::new(
+            &term,
+            &mut search,
+            Point::new(Line(0), Column(1))..=Point::new(Line(0), Column(1)),
+        )
+        .take(1)
+        .count();
+
+        assert_eq!(count, 0);
+    }
+
+    #[test]
+    fn collect_unique_hyperlinks() {
+        let mut term = mock_term("000\r\n111");
+        term.goto(0, 0);
+
+        let hyperlink_foo = Hyperlink::new(Some("1"), String::from("foo"));
+        let hyperlink_bar = Hyperlink::new(Some("2"), String::from("bar"));
+
+        // Create 2 hyperlinks on the first line.
+        term.set_hyperlink(Some(hyperlink_foo.clone().into()));
+        term.input('b');
+        term.input('a');
+        term.set_hyperlink(Some(hyperlink_bar.clone().into()));
+        term.input('r');
+        term.set_hyperlink(Some(hyperlink_foo.clone().into()));
+        term.goto(1, 0);
+
+        // Ditto for the second line.
+        term.set_hyperlink(Some(hyperlink_foo.into()));
+        term.input('b');
+        term.input('a');
+        term.set_hyperlink(Some(hyperlink_bar.into()));
+        term.input('r');
+        term.set_hyperlink(None);
+
+        let mut unique_hyperlinks = visible_unique_hyperlinks_iter(&term);
+        assert_eq!(
+            Some(Match::new(Point::new(Line(0), Column(0)), Point::new(Line(0), Column(1)))),
+            unique_hyperlinks.next()
+        );
+        assert_eq!(
+            Some(Match::new(Point::new(Line(0), Column(2)), Point::new(Line(0), Column(2)))),
+            unique_hyperlinks.next()
+        );
+        assert_eq!(None, unique_hyperlinks.next());
+    }
+
+    #[test]
+    fn visible_regex_match_covers_entire_viewport() {
+        let content = "I'm a match!\r\n".repeat(4096);
+        // The Term returned from this call will have a viewport starting at 0 and ending at 4096.
+        // That's good enough for this test, since it only cares about visible content.
+        let term = mock_term(&content);
+        let mut regex = RegexSearch::new("match!").unwrap();
+
+        // The iterator should match everything in the viewport.
+        assert_eq!(visible_regex_match_iter(&term, &mut regex).count(), 4096);
+    }
 }
